#define MODULE "tmpfs"

#include "fs/TmpFS.h"
#include "log/Log.h"
#include "memory/MemoryManager.h"
#include "misc/utils.h"
#include "std/errno.h"
#include "std/string.h"
#include "utils/Dynamic.h"
#include "utils/move.h"

namespace TmpFS
{
    struct File
    {
        void* ptr;
        uint64_t pages;
        uint64_t length;
    };

    struct Directory
    {
        Dynamic<VFS::Node*> files;
    };
}

VFS::Node* tmpfs_root = nullptr;

Dynamic<TmpFS::Directory> tmpfs_dirs;
Dynamic<TmpFS::File> tmpfs_files;

extern uint64_t clock_now();

static bool resize_file(TmpFS::File& file, size_t new_size)
{
    size_t pages = Utilities::get_blocks_from_size(PAGE_SIZE, new_size);
    if (pages > file.pages) // resize only if we need a bigger chunk of memory
    {
        auto rc = MemoryManager::get_pages(pages, MAP_READ_WRITE);
        if (rc.has_error()) return false;
        void* new_ptr = rc.release_value();
        if (file.ptr)
        {
            memcpy(new_ptr, file.ptr, file.length);
            MemoryManager::release_pages(file.ptr, file.pages);
        }
        file.pages = pages;
        file.ptr = new_ptr;
    }
    file.length = new_size;
    return true;
}

VFS::Node* TmpFS::get()
{
    if (tmpfs_root) return tmpfs_root;
    kinfoln("initializing tmpfs instance");
    tmpfs_root = new VFS::Node;
    tmpfs_root->length = 0;
    tmpfs_root->inode = 0;
    tmpfs_root->type = VFS_DIRECTORY;
    tmpfs_root->find_func = TmpFS::finddir;
    tmpfs_root->readdir_func = TmpFS::readdir;
    tmpfs_root->mkdir_func = TmpFS::mkdir;
    tmpfs_root->create_func = TmpFS::create;
    tmpfs_root->mode = 0777;
    tmpfs_root->uid = tmpfs_root->gid = 0;
    tmpfs_root->atime = tmpfs_root->ctime = tmpfs_root->mtime = clock_now();
    strncpy(tmpfs_root->name, "tmpfs", sizeof(tmpfs_root->name));
    tmpfs_dirs.ensure_capacity(16);
    tmpfs_files.ensure_capacity(16);
    tmpfs_dirs.set_expand_rate(16);
    tmpfs_files.set_expand_rate(16);
    ensure(tmpfs_dirs.push({{}}));
    return tmpfs_root;
}

VFS::Node* TmpFS::finddir(VFS::Node* node, const char* filename)
{
    if (!node) return nullptr;
    if (node->inode >= tmpfs_dirs.size()) return nullptr;
    auto& dir = tmpfs_dirs[node->inode];
    kdbgln("searching for '%s' in '%s'", filename, node->name);
    for (size_t i = 0; i < dir.files.size(); i++)
    {
        if (!strncmp(dir.files[i]->name, filename, sizeof(dir.files[i]->name))) return dir.files[i];
    }
    return nullptr;
}

VFS::Node* TmpFS::readdir(VFS::Node* node, long offset)
{
    if (!node) return nullptr;
    if (node->inode >= tmpfs_dirs.size()) return nullptr;
    auto& dir = tmpfs_dirs[node->inode];
    if (offset < 0) return nullptr;
    if ((size_t)offset >= dir.files.size()) return nullptr;
    kdbgln("reading offset %ld in '%s'", offset, node->name);
    return dir.files[offset];
}

int TmpFS::mkdir(VFS::Node* node, const char* name, mode_t mode)
{
    if (node->inode >= tmpfs_dirs.size()) return -EINVAL;
    if (!(node->type & VFS_DIRECTORY)) return -ENOTDIR;
    auto& parent = tmpfs_dirs[node->inode];
    VFS::Node* new_node = new VFS::Node;
    new_node->inode = tmpfs_dirs.size();
    new_node->find_func = TmpFS::finddir;
    new_node->readdir_func = TmpFS::readdir;
    new_node->mkdir_func = TmpFS::mkdir;
    new_node->create_func = TmpFS::create;
    new_node->length = 0;
    new_node->type = VFS_DIRECTORY;
    new_node->mode = mode;
    new_node->uid = new_node->gid = 0;
    new_node->atime = new_node->ctime = new_node->mtime = clock_now();
    strncpy(new_node->name, name, sizeof(new_node->name));
    ensure(parent.files.push(
        new_node)); // We have to push the new node to the parent BEFORE we push the directory object to the list of
                    // tmpfs directories. This is crucial, since pushing to the list of tmpfs directories could relocate
                    // said list, where the parent object is stored, while we would still be using the old parent
                    // object. Thus our changes to the parent would be erased as soon as we exit this function.
    ensure(tmpfs_dirs.push({{}})); // FIXME: Handle OOM instead of panicking
    node->length++;
    kdbgln("created directory '%s' in '%s' (which now has %lu entries), mode %o", name, node->name, node->length, mode);
    return 0;
}

VFS::Node* TmpFS::create(VFS::Node* node, const char* name, mode_t mode, uid_t uid, gid_t gid)
{
    if (node->inode >= tmpfs_dirs.size()) return nullptr;
    if (!(node->type & VFS_DIRECTORY)) return nullptr;
    auto& parent = tmpfs_dirs[node->inode];
    VFS::Node* new_node = new VFS::Node;
    new_node->inode = tmpfs_files.size();
    new_node->length = 0;
    new_node->read_func = TmpFS::read;
    new_node->write_func = TmpFS::write;
    new_node->type = VFS_FILE;
    new_node->mode = mode;
    new_node->uid = uid;
    new_node->gid = gid;
    new_node->atime = new_node->ctime = new_node->mtime = clock_now();
    strncpy(new_node->name, name, sizeof(new_node->name));
    ensure(parent.files.push(
        new_node)); // We have to push the new node to the parent BEFORE we push the directory object to the list of
                    // tmpfs directories. This is crucial, since pushing to the list of tmpfs directories could relocate
                    // said list, where the parent object is stored, while we would still be using the old parent
                    // object. Thus our changes to the parent would be erased as soon as we exit this function.
    ensure(tmpfs_files.push({nullptr, 0, 0})); // FIXME: Handle OOM instead of panicking
    node->length++;
    kdbgln("created file '%s' in '%s' (which now has %lu entries), mode %o", name, node->name, node->length, mode);
    return 0;
}

ssize_t TmpFS::read(VFS::Node* node, size_t offset, size_t length, char* buffer)
{
    kdbgln("got to read()");
    if (!node) return -1;
    if (node->inode >= tmpfs_files.size()) return -1;
    File& file = tmpfs_files[node->inode];
    size_t old_len = length;
    if (offset > file.length) return -1;
    if (offset + length > file.length) { length = file.length - offset; }
    kdbgln("reading %zu bytes from '%s' at offset %zu (requested length was %zu)", length, node->name, offset, old_len);
    memcpy(buffer, (void*)((uintptr_t)file.ptr + offset), length);
    return length;
}

ssize_t TmpFS::write(VFS::Node* node, size_t offset, size_t length, const char* buffer)
{
    if (!node) return -1;
    if (node->inode >= tmpfs_files.size()) return -1;
    File& file = tmpfs_files[node->inode];
    size_t new_length = file.length;
    if (offset + length > file.length) new_length = offset + length;
    if (new_length != file.length)
    {
        if (!resize_file(file, new_length)) return -ENOMEM;
        node->length = new_length;
    }
    kdbgln("writing %zu bytes to '%s' at offset %zu", length, node->name, offset);
    memcpy((void*)((uintptr_t)file.ptr + offset), buffer, length);
    return length;
}